/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: mmaps.c,v 1.1 2007/02/08 06:17:49 phargrov Exp $
 *
 * Test of various mmap()s across a checkpoint
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/mman.h>

#include "crut.h"
#include "crut_util.h"

#define FILENAME "tstfile.0"

static const char MSG[] = "Hello world!\n";
static int pagesize = -1;

static int
mmaps_setup_common(void **testdata, void *start, int prot , int flags, int fd, off_t offset, int copy)
{
    void *p = mmap(start, pagesize, prot, flags, fd, offset);
    CRUT_DEBUG("mmap(%p, %d, 0x%x, 0x%x, %d, %d) -> %p (%d)", start, pagesize, prot, flags, fd, (int)offset, p, errno);
    if (p == MAP_FAILED) return -1;
    if (copy) strcpy(p, MSG);
    *testdata = p;
    return 0;
}

static int mmaps_setup_anon_common(void **testdata, int prot, int flags)
{
    return mmaps_setup_common(testdata, 0, prot, MAP_ANONYMOUS|flags, 0, 0, 1);
}

static int mmaps_setup_file_common(void **testdata, int prot, int flags, int do_unlink)
{
    int fd, retval;
    (void)unlink(FILENAME);
    fd = open(FILENAME, O_CREAT|O_RDWR|O_TRUNC, S_IREAD|S_IWRITE);
    if (fd < 0) return fd;
    write(fd, MSG, sizeof(MSG));
    retval = mmaps_setup_common(testdata, 0, prot, flags, fd, 0, 0);
    close(fd);
    if (do_unlink) (void)unlink(FILENAME);
    return retval;
}

static int mmaps_setup_rw_as(void **testdata)
{ return mmaps_setup_anon_common(testdata, PROT_READ|PROT_WRITE, MAP_SHARED); }

static int mmaps_setup_rw_ap(void **testdata)
{ return mmaps_setup_anon_common(testdata, PROT_READ|PROT_WRITE, MAP_PRIVATE); }

static int mmaps_setup_rw_fs(void **testdata)
{ return mmaps_setup_file_common(testdata, PROT_READ|PROT_WRITE, MAP_SHARED, 0); }

static int mmaps_setup_rw_fp(void **testdata)
{ return mmaps_setup_file_common(testdata, PROT_READ|PROT_WRITE, MAP_PRIVATE, 0); }

static int mmaps_setup_rw_us(void **testdata)
{ return mmaps_setup_file_common(testdata, PROT_READ|PROT_WRITE, MAP_SHARED, 1); }

static int mmaps_setup_rw_up(void **testdata)
{ return mmaps_setup_file_common(testdata, PROT_READ|PROT_WRITE, MAP_PRIVATE, 1); }


static int
mmaps_precheckpoint(void *p)
{
    return 0;
}

static int
mmaps_continue(void *p)
{
    int retval;

    CRUT_DEBUG("Continuing after checkpoint.");
    retval = strncmp((char *)p, MSG, sizeof(MSG)) ? -1 : 0;

    return retval;
}

static int
mmaps_restart(void *p)
{
    int retval;

    CRUT_DEBUG("Restarting from checkpoint.");
    retval = strncmp((char *)p, MSG, sizeof(MSG)) ? -1 : 0;

    return retval;
}

static int
mmaps_teardown(void *p)
{
    (void)unlink(FILENAME);
    return munmap(p, getpagesize());
}

int
main(int argc, char *argv[])
{
    int ret;
    struct crut_operations rw_as_test_ops = {
	test_name:"mmaps_rw_as",
        test_description:"Test mmap(READ|WRITE,ANONYMOUS|SHARED).",
	test_setup:mmaps_setup_rw_as,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };
    struct crut_operations rw_ap_test_ops = {
	test_name:"mmaps_rw_ap",
        test_description:"Test mmap(READ|WRITE,ANONYMOUS|PRIVATE).",
	test_setup:mmaps_setup_rw_ap,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };

    struct crut_operations rw_fs_test_ops = {
	test_name:"mmaps_rw_fs",
        test_description:"Test mmap(READ|WRITE,SHARED) of a file.",
	test_setup:mmaps_setup_rw_fs,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };
    struct crut_operations rw_fp_test_ops = {
	test_name:"mmaps_rw_fp",
        test_description:"Test mmap(READ|WRITE,PRIVATE) of a file.",
	test_setup:mmaps_setup_rw_fp,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };

    struct crut_operations rw_us_test_ops = {
	test_name:"mmaps_rw_us",
        test_description:"Test mmap(READ|WRITE,SHARED) of an unlinked file.",
	test_setup:mmaps_setup_rw_us,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };
    struct crut_operations rw_up_test_ops = {
	test_name:"mmaps_rw_up",
        test_description:"Test mmap(READ|WRITE,PRIVATE) of an unlinked file.",
	test_setup:mmaps_setup_rw_up,
	test_precheckpoint:mmaps_precheckpoint,
	test_continue:mmaps_continue,
	test_restart:mmaps_restart,
	test_teardown:mmaps_teardown,
    };

    /* add the basic tests */
    crut_add_test(&rw_as_test_ops);
    crut_add_test(&rw_ap_test_ops);
    crut_add_test(&rw_fs_test_ops);
    crut_add_test(&rw_fp_test_ops);
    crut_add_test(&rw_us_test_ops);
    crut_add_test(&rw_up_test_ops);

    pagesize = getpagesize();
    ret = crut_main(CR_SCOPE_TREE, argc, argv);

    return ret;
}
